# Copyright 2025 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from requests import Response
from pandas import DataFrame, Series
import pandas as pd
from typing import Union, Optional, List
from spgci.api_client import get_data, Paginator
from spgci.utilities import list_to_filter
from enum import Enum
from datetime import datetime, date
import warnings

class NANaturalGasAnalytics:
    """
    North America Natural Gas Pipeline

    .. deprecated:: 0.0.60
       Use `AmericasGas` instead.
    """

    def __init__(self):
        warnings.warn(
            "NANaturalGasAnalytics is deprecated as of 0.0.60. Use AmericasGas instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    _path = "analytics/natural-gas/north-america/supply-demand/v1/"

    @staticmethod
    def _convert_to_df(resp: Response) -> pd.DataFrame:
        j = resp.json()
        # df = pd.json_normalize(j["results"], record_path=["data"], meta="symbol")  # type: ignore
        df = pd.json_normalize(j["results"])  # type: ignore

        if "gasDate" in df.columns:
            df["gasDate"] = pd.to_datetime(df["gasDate"])  # type: ignore
        if "validFrom" in df.columns:
            df["validFrom"] = pd.to_datetime(df["validFrom"])  # type: ignore
        if "validTo" in df.columns:
            df["validTo"] = pd.to_datetime(df["validTo"], errors="ignore")  # type: ignore
        if "createDate" in df.columns:
            df["createDate"] = pd.to_datetime(df["createDate"])  # type: ignore
        if "modifiedDate" in df.columns:
            df["modifiedDate"] = pd.to_datetime(df["modifiedDate"])  # type: ignore

        return df

    class PointType(Enum):
        """Point Type"""

        Point = "point"
        Segment = "segment"

    class FacilityType(Enum):
        """Facility Type"""

        LDC = "LDC"
        Constraint = "Constraint"
        Compressor = "Compressor"
        Production = "Production"
        GatheringCompany = "Gathering Company"
        Interconnect = "Interconnect"
        Lateral = "Lateral"
        ProcessingPlant = "Processing Plant"
        LNG = "LNG"
        End_User = "End User"
        StorageParkLoan = "Storage Park/Loan"
        IntraStateConnect = "IntraState Connect"
        GasTreatmentPlants = "Gas Treatment Plants"
        Demand = "Demand"
        Municipal = "Municipal"
        PowerPlant = "Power Plant"
        PipelineAdministration = "Pipeline Administration"
        FieldFuel = "Field Fuel"
        EnhancedOilRecoveryCogen = "Enhanced Oil Recovery/Cogen"

    @staticmethod
    def _paginate(resp: Response) -> Paginator:
        j = resp.json()
        count = j["metadata"]["count"]
        size = j["metadata"]["pageSize"]

        remainder = count % size
        quotient = count // size
        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages=total_pages)

    def get_pipeline_flows(
        self,
        *,
        pipeline_id: Optional[int] = None,
        component_id: Optional[int] = None,
        gasdate: Optional[Union[date, list[date], "Series[date]"]] = None,
        gasdate_gt: Optional[date] = None,
        gasdate_gte: Optional[date] = None,
        gasdate_lt: Optional[date] = None,
        gasdate_lte: Optional[date] = None,
        nomination_cycle: Optional[Union[str, list[str], "Series[str]"]] = None,
        location_type: Optional[Union[str, list[str], "Series[str]"]] = None,
        scheduled_volume: Optional[float] = None,
        actual_volume: Optional[float] = None,
        utilization: Optional[float] = None,
        design_capacity: Optional[float] = None,
        operational_capacity: Optional[float] = None,
        actual_capacity: Optional[float] = None,
        operationally_available: Optional[float] = None,
        flow_dir: Optional[Union[str, list[str], "Series[str]"]] = None,
        interruptible_flow: Optional[Union[str, list[str], "Series[str]"]] = None,
        valid_from: Optional[datetime] = None,
        valid_to: Optional[datetime] = None,
        data_active: Optional[bool] = None,
        data_source: Optional[str] = None,
        create_date: Optional[datetime] = None,
        modified_date: Optional[datetime] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 2000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[Response, DataFrame]:
        """
        Analytics data for North America Natural Gas Pipelines .

        Parameters
        ----------
        pipeline_id : int, optional
            filter by pipe_line_id, by default None
        component_id : int, optional
            filter by component_id, by default None
        gasdate : Optional[Union[str, list[datetime], Series[datetime]]], optional
            filter by gasdate, by default None
        gasdate_gt : Optional[datetime], optional
            filter by ``gasdate > x``, by default None
        gasdate_gte : Optional[datetime], optional
            filter by ``gasdate >= x``, by default None
        gasdate_lt : Optional[datetime], optional
            filter by ``gasdate < x``, by default None
        gasdate_lte : Optional[datetime], optional
            filter by ``gasdate <= x``, by default None
        nomination_cycle: Optional[Union[str, list[str], "Series[str]"]] , optional
            filter by nomination_cycle, by default None
        location_type: Optional[Union[str, list[str], "Series[str]"]] , optional
            filter by location_type, by default None
        scheduled_volume: Optional[float] , optional
            filter by scheduled_volume, by default None
        actual_volume: Optional[float] , optional
            filter by actual_volume, by default None
        utilization: Optional[float] , optional
            filter by utilization, by default None
        design_capacity: Optional[float] , optional
            filter by design_capacity, by default None
        operational_capacity: Optional[float] , optional
            filter by operational_capacity, by default None
        actual_capacity: Optional[float] , optional
            filter by actual_capacity, by default None
        operationally_available: Optional[float] , optional
            filter by operationally_available, by default None,
        flow_dir: Optional[Union[str, list[str], "Series[str]"]] , optional
            filter by flow_dir, by default None
        interruptible_flow: Optional[Union[str, list[str], "Series[str]"]] , optional
            filter by interruptible_flow, by default None
        valid_from: Optional[datetime] , optional
            filter by valid_from, by default None
        valid_to: Optional[datetime], optional
            filter by valid_to, by default None
        data_active: Optional[bool] , optional
            filter by data_active, by default None
        data_source: Optional[str], optional
            filter by data_source, by default None
        create_date: Optional[datetime], optional
            filter by create_date, by default None
        modified_date: Optional[datetime], optional
            filter by modified_date, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 2000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
              DataFrame
                  DataFrame of the ``response.json()``
              Response
                  Raw ``requests.Response`` object
        Examples
        --------
        **Get Pipeline flow data by date and nomination cycle**
        >>> d = date(2023, 7, 1)
        >>> ci.NANaturalGasAnalytics().get_pipeline_flows(gasdate=d, nomination_cycle="I2")

        **Using Enum**
        >>> ng = ci.NANaturalGasAnalytics().get_pipeline_flows(faciltyType=ci.NANaturalGasAnalytics().FacilityType.LNG)
        """
        endpoint_path = "pipeline-flow-data"

        filter_param: List[str] = []

        filter_param.append(list_to_filter("pipelineId", pipeline_id))
        filter_param.append(list_to_filter("componentId", component_id))
        filter_param.append(list_to_filter("gasDate", gasdate))
        filter_param.append(list_to_filter("nominationCycle", nomination_cycle))
        filter_param.append(list_to_filter("locationType", location_type))
        filter_param.append(list_to_filter("scheduledVolume", scheduled_volume))
        filter_param.append(list_to_filter("actualVolume", actual_volume))
        filter_param.append(list_to_filter("utilization", utilization))
        filter_param.append(list_to_filter("designCapacity", design_capacity))
        filter_param.append(list_to_filter("operationalCapacity", operational_capacity))
        filter_param.append(list_to_filter("actualCapacity", actual_capacity))
        filter_param.append(
            list_to_filter("operationallyAvailable", operationally_available)
        )
        filter_param.append(list_to_filter("flowDir", flow_dir))
        filter_param.append(list_to_filter("interruptibleFlow", interruptible_flow))
        filter_param.append(list_to_filter("validFrom", valid_from))
        filter_param.append(list_to_filter("validTo", valid_to))
        filter_param.append(list_to_filter("dataActive", data_active))
        filter_param.append(list_to_filter("dataSource", data_source))
        filter_param.append(list_to_filter("createDate", create_date))
        filter_param.append(list_to_filter("modifiedDate", modified_date))

        if gasdate_gt:
            filter_param.append(f'gasDate > "{gasdate_gt}"')
        if gasdate_gte:
            filter_param.append(f'gasDate >= "{gasdate_gte}"')
        if gasdate_lt:
            filter_param.append(f'gasDate < "{gasdate_lt}"')
        if gasdate_lte:
            filter_param.append(f'gasDate <= "{gasdate_lte}"')

        filter_param = [fp for fp in filter_param if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_param)
        elif len(filter_param) > 0:
            filter_exp = " AND ".join(filter_param) + " AND (" + filter_exp + ")"

        params = {
            "pageSize": page_size,
            "filter": filter_exp,
            "page": page,
        }
        return get_data(
            path=f"{self._path}{endpoint_path}",
            df_fn=self._convert_to_df,
            params=params,
            paginate=paginate,
            paginate_fn=self._paginate,
            raw=raw,
        )

    @staticmethod
    def _search_to_df(resp: Response) -> pd.DataFrame:
        j = resp.json()
        df = pd.DataFrame(j["results"])

        # if len(df) > 0:
        #     cols = ["symbol", "description"]
        #     df: pd.DataFrame = df[cols + [x for x in df.columns if x not in cols]]  # type: ignore

        return df

    @staticmethod
    def _ref_paginate(resp: Response) -> Paginator:
        j = resp.json()
        count = j["metadata"]["count"]
        size = j["metadata"]["pageSize"]

        remainder = count % size
        quotient = count // size
        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages=total_pages)

    def get_pipelines(
        self,
        *,
        q: Optional[str] = None,
        pipeline_id: Optional[Union[list[int], "Series[int]", int]] = None,
        pipeline_name: Optional[Union[list[str], "Series[str]", str]] = None,
        point_type: Optional[Union[list[str], "Series[str]", str]] = None,
        point_name: Optional[Union[list[str], "Series[str]", str]] = None,
        drn_number: Optional[Union[list[str], "Series[str]", str]] = None,
        county: Optional[Union[list[str], "Series[str]", str]] = None,
        state: Optional[Union[list[str], "Series[str]", str]] = None,
        country: Optional[Union[list[str], "Series[str]", str]] = None,
        connecting_party_name: Optional[Union[list[str], "Series[str]", str]] = None,
        facility_type: Optional[Union[list[str], "Series[str]", str]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 2000,
        paginate: bool = False,
        raw: bool = False,
    ) -> Union[pd.DataFrame, Response]:
        """
        Fetch Pipelines from the NANaturalGasPipeline API.

        Parameters
        ----------
        q : Optional[str], optional
            filter across fields using free text search, by default None
        pipeline_id : Optional[Union[list[int], Series[int], int]], optional
            filter by pipelineId, by default None
        pipeline_name : Optional[Union[list[str], Series[str], str]], optional
            filter by pipelineName, by default None
        point_type : Optional[Union[list[str], Series[str], str]], optional
            filter by pointName, by default None
        point_name : Optional[Union[list[str], Series[str], str]], optional
            filter by pointName, by default None
        drn_number : Optional[Union[list[str], Series[str], str]], optional
            filter by drnNumber, by default None
        county : Optional[Union[list[str], Series[str], str]], optional
            filter by county, by default None
        state : Optional[Union[list[str], Series[str], str]], optional
            filter by state, by default None
        country : Optional[Union[list[str], Series[str], str]], optional
            filter by country, by default None
        connecting_party_name : Optional[Union[list[str], Series[str], str]], optional
            filter by connectingPartyName, by default None
        facility_type : Optional[Union[list[str], Series[str], str]], optional
            filter by facilityType, by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 2000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object
        Examples
        --------
        **Using String**
        >>> ci.NANaturalGasAnalytics().get_pipelines(point_type="Segment")

        **Using List**
        >>> ci.NANaturalGasAnalytics().get_pipelines(state=["WA", "CO"])

        """
        endpoint_path = "pipeline-reference-data"

        filter_params: List[str] = []

        filter_params.append(list_to_filter("pipelineName", pipeline_name))
        filter_params.append(list_to_filter("pipelineId", pipeline_id))
        filter_params.append(list_to_filter("pointType", point_type))
        filter_params.append(list_to_filter("pointName", point_name))
        filter_params.append(list_to_filter("drnNumber", drn_number))
        filter_params.append(list_to_filter("county", county))
        filter_params.append(list_to_filter("state", state))
        filter_params.append(list_to_filter("country", country))
        filter_params.append(
            list_to_filter("connectingPartyName", connecting_party_name)
        )
        filter_params.append(list_to_filter("facilityType", facility_type))

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {
            "q": q,
            "filter": filter_exp,
            "page": page,
            "pageSize": page_size,
        }
        return get_data(
            path=f"{self._path}{endpoint_path}",
            df_fn=self._search_to_df,
            paginate_fn=self._ref_paginate,
            params=params,
            paginate=paginate,
            raw=raw,
        )



# Copyright 2025 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Union, Optional, List
from requests import Response
from spgci.api_client import get_data, Paginator
from spgci.utilities import list_to_filter, convert_date_to_filter_exp
from pandas import Series, DataFrame, to_datetime  # type: ignore
from datetime import date


class StructuredHeards:
    """
    StructuredHeards Data

    Includes
    --------

    ``get_heards()`` to fetch structured heards.\n
    ``get_markets()`` to fetch market and attributes associated with it.\n

    """

    _endpoint = "structured-heards/v1/"

    @staticmethod
    def _paginate(resp: Response) -> Paginator:
        j = resp.json()
        total_pages = j["metadata"]["total_pages"]

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages)

    @staticmethod
    def _convert_to_df(resp: Response) -> DataFrame:
        j = resp.json()
        df = DataFrame(j["results"])

        if len(df) > 0:
            if "updatedDate" in df.columns:
                df["updatedDate"] = to_datetime(df["updatedDate"])
            if "rtpTimestamp" in df.columns:
                df["rtpTimestamp"] = to_datetime(df["rtpTimestamp"])
        return df

    @staticmethod
    def _paginate_outages(resp: Response) -> Paginator:
        j = resp.json()
        count = j["metadata"]["count"]
        size = j["metadata"]["pagesize"]

        remainder = count % size
        quotient = count // size
        total_pages = quotient + (1 if remainder > 0 else 0)

        if total_pages <= 1:
            return Paginator(False, "page", 1)

        return Paginator(True, "page", total_pages=total_pages)

    def get_heards(
        self,
        market: Union[list[str], "Series[str]", str],
        *,
        geography: Optional[Union[list[str], "Series[str]", str]] = None,
        commodity: Optional[Union[list[str], "Series[str]", str]] = None,
        location: Optional[Union[list[str], "Series[str]", str]] = None,
        heard_type: Optional[Union[list[str], "Series[str]", str]] = None,
        updated_date_lt: Optional[date] = None,
        updated_date_lte: Optional[date] = None,
        updated_date_gt: Optional[date] = None,
        updated_date_gte: Optional[date] = None,
        rtp_timestamp_lt: Optional[date] = None,
        rtp_timestamp_lte: Optional[date] = None,
        rtp_timestamp_gt: Optional[date] = None,
        rtp_timestamp_gte: Optional[date] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        raw: bool = False,
        paginate: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Fetch the data based on the filter expression.

        Parameters
        ----------
        market : Optional[Union[list[str], Series[str], str]]
            filter by market
        geography : Optional[Union[list[str], Series[str], str]], optional
            filter by geography, by default None
        commodity : Optional[Union[list[str], Series[str], str]], optional
            filter by commodity, by default None
        location : Optional[Union[list[str], Series[str], str]], optional
            filter by location, by default None
        heard_type : Optional[Union[list[str], Series[str], str]], optional
            filter by heard_type, by default None
        updated_date_lt: Optional[date], optional
            filter by ``updatedDate < x``, by default None
        updated_date_lte : Optional[date], optional
            filter by ``updatedDate <= x``, by default None
        updated_date_gt : Optional[date], optional
            filter by ``updatedDate > x``, by default None
        updated_date_gte : Optional[date], optional
            filter by ``updatedDate >= x``, by default None
        rtp_timestamp_lt : Optional[date], optional
            filter by ``rtpTimestamp < x``, by default None
        rtp_timestamp_lte : Optional[date], optional
            filter by ``rtpTimestamp <= x``, by default None
        rtp_timestamp_gt : Optional[date], optional
            filter by ``rtpTimestamp > x``, by default None
        rtp_timestamp_gte : Optional[date], optional
            filter by ``rtpTimestamp >= x``, by default None
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False
        filter_exp: string
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.StructuredHeards().get_heards(market="Americas crude oil")
        """
        endpoint_path = "data"
        filter_params: List[str] = []
        filter_params.append(list_to_filter("geography", geography))
        filter_params.append(list_to_filter("commodity", commodity))
        filter_params.append(list_to_filter("location", location))
        filter_params.append(list_to_filter("market", market))
        filter_params.append(list_to_filter("heard_type", heard_type))

        filter_params = convert_date_to_filter_exp(
            "updatedDate",
            updated_date_gt,
            updated_date_gte,
            updated_date_lt,
            updated_date_lte,
            filter_params,
        )
        filter_params = convert_date_to_filter_exp(
            "rtpTimeStamp",
            rtp_timestamp_gt,
            rtp_timestamp_gte,
            rtp_timestamp_lt,
            rtp_timestamp_lte,
            filter_params,
        )

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"page": page, "pageSize": page_size, "filter": filter_exp}

        response = get_data(
            path=f"{self._endpoint}{endpoint_path}",
            params=params,
            df_fn=self._convert_to_df,
            paginate_fn=self._paginate,
            raw=raw,
            paginate=paginate,
        )
        return response

    def get_markets(
        self,
        *,
        market: Optional[Union[list[str], "Series[str]", str]] = None,
        attributes: Optional[Union[list[str], "Series[str]", str]] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        raw: bool = False,
        paginate: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Fetch the data based on the filter expression.

        Parameters
        ----------
        market : Optional[Union[list[str], Series[str], str]], optional
            filter by commodity_name, by default None
        attributes : Optional[Union[list[str], Series[str], str]], optional
            filter by train_name, by default None
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False
        filter_exp: string
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.StructuredHeards().get_markets()
        """
        endpoint_path = "markets"
        filter_params: List[str] = []
        filter_params.append(list_to_filter("market", market))
        filter_params.append(list_to_filter("attributes", attributes))

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"page": page, "pageSize": page_size, "filter": filter_exp}

        response = get_data(
            path=f"{self._endpoint}{endpoint_path}",
            params=params,
            df_fn=self._convert_to_df,
            paginate_fn=self._paginate_outages,
            raw=raw,
            paginate=paginate,
        )
        return response



# Copyright 2025 S&P Global Commodity Insights

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#       http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import List, Optional, Union
from requests import Response
from spgci.api_client import get_data
from spgci.utilities import list_to_filter
from pandas import DataFrame, Series
from datetime import date, datetime
from packaging.version import parse
import pandas as pd


class Weather:
    def get_actual(
        self,
        market: Optional[Union[list[str], Series[str], str]] = None,
        city: Optional[Union[list[str], Series[str], str]] = None,
        location: Optional[Union[list[str], Series[str], str]] = None,
        weather_date: Optional[Union[list[date], Series[date], date]] = None,
        weather_date_gte: Optional[date] = None,
        weather_date_gt: Optional[date] = None,
        weather_date_lte: Optional[date] = None,
        weather_date_lt: Optional[date] = None,
        modified_date: Optional[
            Union[list[datetime], Series[datetime], datetime]
        ] = None,
        modified_date_gte: Optional[datetime] = None,
        modified_date_gt: Optional[datetime] = None,
        modified_date_lte: Optional[datetime] = None,
        modified_date_lt: Optional[datetime] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        raw: bool = False,
        paginate: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Fetch the data based on the filter expression.

        Parameters
        ----------

        market: Optional[Union[list[str], Series[str], str]]
            The market in which the weather data is recorded or observed., be default None
        city: Optional[Union[list[str], Series[str], str]]
            The specific city within the country for which the weather data is being reported., be default None
        location: Optional[Union[list[str], Series[str], str]]
            Weather station which provides Differentiator between country weather data and city weather data., be default None
        weather_date: Optional[Union[list[date], Series[date], date]]
            The date for which the weather data is recorded or forecasted., be default None
        weather_date_gte: Optional[date]
            filter by ``weatherDate >= x`` , by default None
        weather_date_gt: Optional[date]
            filter by ``weatherDate > x`` , by default None
        weather_date_lte: Optional[date]
            filter by ``weatherDate <= x`` , by default None
        weather_date_lt: Optional[date]
            filter by ``weatherDate < x`` , by default None
        modified_date: Optional[Union[list[datetime], Series[datetime], datetime]]
            The latest date of modification for the Weather Actual., be default None
        modified_date_gte: Optional[datetime]
            filter by ``modifiedDate >= x`` , by default None
        modified_date_gt: Optional[datetime]
            filter by ``modifiedDate > x`` , by default None
        modified_date_lte: Optional[datetime]
            filter by ``modifiedDate <= x`` , by default None
        modified_date_lt: Optional[datetime]
            filter by ``modifiedDate < x`` , by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.Weather().get_actual(city="Boston")

        **Date Range**
        >>> ci.Weather().get_actual(market="United States", weather_date_gte="2024-01-01", weather_date_lte="2024-01-31")
        """

        filter_params: List[str] = []
        filter_params.append(list_to_filter("market", market))
        filter_params.append(list_to_filter("city", city))
        filter_params.append(list_to_filter("location", location))
        filter_params.append(list_to_filter("weatherDate", weather_date))
        filter_params.append(list_to_filter("modifiedDate", modified_date))

        if weather_date_gt is not None:
            filter_params.append(f'weatherDate > "{weather_date_gt}"')
        if weather_date_gte is not None:
            filter_params.append(f'weatherDate >= "{weather_date_gte}"')
        if weather_date_lt is not None:
            filter_params.append(f'weatherDate < "{weather_date_lt}"')
        if weather_date_lte is not None:
            filter_params.append(f'weatherDate <= "{weather_date_lte}"')

        if modified_date_gt is not None:
            filter_params.append(f'modifiedDate > "{modified_date_gt}"')
        if modified_date_gte is not None:
            filter_params.append(f'modifiedDate >= "{modified_date_gte}"')
        if modified_date_lt is not None:
            filter_params.append(f'modifiedDate < "{modified_date_lt}"')
        if modified_date_lte is not None:
            filter_params.append(f'modifiedDate <= "{modified_date_lte}"')

        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"page": page, "pageSize": page_size, "filter": filter_exp}

        response = get_data(
            path="/weather/v1/actual",
            params=params,
            df_fn=self._convert_to_df,
            raw=raw,
            paginate=paginate,
        )
        return response

    def get_forecast(
        self,
        market: Optional[Union[list[str], Series[str], str]] = None,
        city: Optional[Union[list[str], Series[str], str]] = None,
        location: Optional[Union[list[str], Series[str], str]] = None,
        recorded_date: Optional[Union[list[date], Series[date], date]] = None,
        recorded_date_gte: Optional[date] = None,
        recorded_date_gt: Optional[date] = None,
        recorded_date_lte: Optional[date] = None,
        recorded_date_lt: Optional[date] = None,
        weather_date: Optional[Union[list[date], Series[date], date]] = None,
        weather_date_gte: Optional[date] = None,
        weather_date_gt: Optional[date] = None,
        weather_date_lte: Optional[date] = None,
        weather_date_lt: Optional[date] = None,
        modified_date: Optional[
            Union[list[datetime], Series[datetime], datetime]
        ] = None,
        modified_date_gte: Optional[datetime] = None,
        modified_date_gt: Optional[datetime] = None,
        modified_date_lte: Optional[datetime] = None,
        modified_date_lt: Optional[datetime] = None,
        filter_exp: Optional[str] = None,
        page: int = 1,
        page_size: int = 1000,
        raw: bool = False,
        paginate: bool = False,
    ) -> Union[DataFrame, Response]:
        """
        Fetch the data based on the filter expression.

        Parameters
        ----------

        market: Optional[Union[list[str], Series[str], str]]
            The market in which the weather data is recorded or observed., be default None
        city: Optional[Union[list[str], Series[str], str]]
            The specific city within the country for which the weather data is being reported., be default None
        location: Optional[Union[list[str], Series[str], str]]
            Weather station which provides Differentiator between country weather data and city weather data., be default None
        recorded_date: Optional[Union[list[date], Series[date], date]]
            Recorded Date refers to the specific date on which weather updateds are recorded., be default None
        recorded_date_gte: Optional[date]
            filter by ``recordedDate >= x`` , by default None
        recorded_date_gt: Optional[date]
            filter by ``recordedDate > x`` , by default None
        recorded_date_lte: Optional[date]
            filter by ``recordedDate <= x`` , by default None
        recorded_date_lt: Optional[date]
            filter by ``recordedDate < x`` , by default None
        weather_date: Optional[Union[list[date], Series[date], date]]
            The date for which the weather data is recorded or forecasted., be default None
        weather_date_gte: Optional[date]
            filter by ``weatherDate >= x`` , by default None
        weather_date_gt: Optional[date]
            filter by ``weatherDate > x`` , by default None
        weather_date_lte: Optional[date]
            filter by ``weatherDate <= x`` , by default None
        weather_date_lt: Optional[date]
            filter by ``weatherDate < x`` , by default None
        modified_date: Optional[datetime]
            The latest date of modification for the Weather Forecast., be default None
        modified_date_gte: Optional[datetime]
            filter by ``modifiedDate >= x`` , by default None
        modified_date_gt: Optional[datetime]
            filter by ``modifiedDate > x`` , by default None
        modified_date_lte: Optional[datetime]
            filter by ``modifiedDate <= x`` , by default None
        modified_date_lt: Optional[datetime]
            filter by ``modifiedDate < x`` , by default None
        filter_exp : Optional[str], optional
            pass-thru ``filter`` query param to use a handcrafted filter expression, by default None
        page : int, optional
            pass-thru ``page`` query param to request a particular page of results, by default 1
        page_size : int, optional
            pass-thru ``pageSize`` query param to request a particular page size, by default 1000
        paginate : bool, optional
            whether to auto-paginate the response, by default False
        raw : bool, optional
            return a ``requests.Response`` instead of a ``DataFrame``, by default False

        Returns
        -------
        Union[pd.DataFrame, Response]
            DataFrame
                DataFrame of the ``response.json()``
            Response
                Raw ``requests.Response`` object

        Examples
        --------
        **Simple**
        >>> ci.Weather().get_forecast(city="Boston")

        **Date Range**
        >>> ci.Weather().get_forecast(market="United States", weather_date_gte="2024-01-01", weather_date_lte="2024-01-31")

        """

        filter_params: List[str] = []
        filter_params.append(list_to_filter("market", market))
        filter_params.append(list_to_filter("city", city))
        filter_params.append(list_to_filter("location", location))
        filter_params.append(list_to_filter("recordedDate", recorded_date))
        filter_params.append(list_to_filter("weatherDate", weather_date))
        filter_params.append(list_to_filter("modifiedDate", modified_date))

        if recorded_date_gt is not None:
            filter_params.append(f'recordedDate > "{recorded_date_gt}"')
        if recorded_date_gte is not None:
            filter_params.append(f'recordedDate >= "{recorded_date_gte}"')
        if recorded_date_lt is not None:
            filter_params.append(f'recordedDate < "{recorded_date_lt}"')
        if recorded_date_lte is not None:
            filter_params.append(f'recordedDate <= "{recorded_date_lte}"')

        if weather_date_gt is not None:
            filter_params.append(f'weatherDate > "{weather_date_gt}"')
        if weather_date_gte is not None:
            filter_params.append(f'weatherDate >= "{weather_date_gte}"')
        if weather_date_lt is not None:
            filter_params.append(f'weatherDate < "{weather_date_lt}"')
        if weather_date_lte is not None:
            filter_params.append(f'weatherDate <= "{weather_date_lte}"')

        if modified_date_gt is not None:
            filter_params.append(f'modifiedDate > "{modified_date_gt}"')
        if modified_date_gte is not None:
            filter_params.append(f'modifiedDate >= "{modified_date_gte}"')
        if modified_date_lt is not None:
            filter_params.append(f'modifiedDate < "{modified_date_lt}"')
        if modified_date_lte is not None:
            filter_params.append(f'modifiedDate <= "{modified_date_lte}"')
        filter_params = [fp for fp in filter_params if fp != ""]

        if filter_exp is None:
            filter_exp = " AND ".join(filter_params)
        elif len(filter_params) > 0:
            filter_exp = " AND ".join(filter_params) + " AND (" + filter_exp + ")"

        params = {"page": page, "pageSize": page_size, "filter": filter_exp}

        response = get_data(
            path="/weather/v1/forecast",
            params=params,
            df_fn=self._convert_to_df,
            raw=raw,
            paginate=paginate,
        )
        return response

    @staticmethod
    def _convert_to_df(resp: Response) -> pd.DataFrame:
        j = resp.json()
        df = pd.json_normalize(j["results"])  # type: ignore

        if "weatherDate" in df.columns:
            if parse(pd.__version__) >= parse("2"):
                df["weatherDate"] = pd.to_datetime(df["weatherDate"], format="ISO8601")
            else:
                df["weatherDate"] = pd.to_datetime(df["weatherDate"])

        if "modifiedDate" in df.columns:
            if parse(pd.__version__) >= parse("2"):
                df["modifiedDate"] = pd.to_datetime(
                    df["modifiedDate"], format="ISO8601"
                )
            else:
                df["modifiedDate"] = pd.to_datetime(df["modifiedDate"])

        if "recordedDate" in df.columns:
            if parse(pd.__version__) >= parse("2"):
                df["recordedDate"] = pd.to_datetime(
                    df["recordedDate"], format="ISO8601"
                )
            else:
                df["recordedDate"] = pd.to_datetime(df["recordedDate"])
        return df
